@@ -1,4 +1,5 @@
 """Runs the tests"""
+import datetime
 import sys
 import os
 import json
@@ -117,7 +118,7 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
             progress_json['overall_result'] = overall_result
 
         test_control['progress'] = progress_json
-        progess_queue.put(json.dumps(progress_json))
+        progess_queue.put(json.dumps(progress_json, default=str))
 
     common_definitions = get_common_definitions()
 
@@ -144,6 +145,8 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
                     dut_status[dut] = {'step': None, 'status': 'wait', 'sn': None}
                     if dut in last_dut_status:
                         dut_status[dut]['test_status'] = last_dut_status[dut]['test_status']
+                    else:
+                        dut_status[dut]['test_status'] = 'idle'
 
                 report_progress("Prepare", duts=dut_status)
 
@@ -158,17 +161,33 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
 
             tests = [t for t in test_definitions.TESTS if t not in test_definitions.SKIP]
 
+            failed_steps = {}
+
             for dut_key, dut_value in duts.items():
-                dut_status[dut_key] = {
-                    'step': None,
-                    'status': 'idle',
-                    'test_status': None,
-                    'sn': dut_value['sn'],
-                }
+                if dut_value:
+                    dut_status[dut_key] = {
+                        'step': None,
+                        'status': 'idle',
+                        'test_status': None,
+                        'sn': dut_value['sn'],
+                    }
+                    failed_steps[dut_key] = {}
+                elif dut_key not in last_dut_status:
+                    dut_status[dut_key] = {
+                        'step': None,
+                        'status': 'idle',
+                        'test_status': None,
+                    }
+
+            results["start_time"] = datetime.datetime.now()
+
+            prev_results = {}
 
             for test_case in tests:
 
                 for dut_name, dut_value in duts.items():
+                    if not dut_value:
+                        continue
                     dut_sn = dut_value['sn']
                     dut_status[dut_name]['step'] = test_case
                     dut_status[dut_name]['status'] = 'testing'
@@ -180,8 +199,14 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
 
                     results[dut_sn]["test_position"] = dut_name
 
+                    start_time = datetime.datetime.now()
+
                     test_instance = getattr(test_definitions, test_case)()
 
+                    if dut_sn in test_instance.previous_results:
+
+                        test_instance.previous_results = prev_results[dut_sn]
+
                     try:
                         test_instance.test(common_definitions.INSTRUMENTS, dut_sn)
                     except Exception as err:
@@ -197,23 +222,36 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
                                 test_definitions.LIMITS[test_case]
                             )
                         else:
+                            # Todo: "no-limit test" not working. Tjeu: Create test single test without limit
                             results[dut_sn][test_case] = test_instance.result_handler(None)
                         # Clean
                         if hasattr(test_instance, 'clean'):
                             test_instance.clean(common_definitions.INSTRUMENTS, dut_sn)
 
+                    prev_results[dut_sn] = test_instance.results
+
                     if not all([r[1]["result"] for r in results[dut_sn][test_case].items()]):
                         overall_result = False
-                        dut_status[dut_name]['test_status'] = 'fail'
-                        dut_status[dut_name]['failed_step'] = test_case
+                        if 'failed_step' in failed_steps[dut_name]:
+                            failed_steps[dut_name]['failed_step'] = (
+                                failed_steps[dut_name]['failed_step'] + ', ' + test_case
+                            )
+                        else:
+                            failed_steps[dut_name]['failed_step'] = test_case
                         if hasattr(test_instance, 'clean_fail'):
                             test_instance.clean_fail(common_definitions.INSTRUMENTS, dut_sn)
 
                     else:
-                        dut_status[dut_name]['test_status'] = 'pass'
                         if hasattr(test_instance, 'clean_pass'):
                             test_instance.clean_pass(common_definitions.INSTRUMENTS, dut_sn)
 
+                    results[dut_sn][test_case]["end_time"] = datetime.datetime.now()
+                    results[dut_sn][test_case]["start_time"] = start_time
+
+                    results[dut_sn][test_case]["duration_s"] = (
+                        results[dut_sn][test_case]["end_time"]
+                        - results[dut_sn][test_case]["start_time"]
+                    ).total_seconds()
 
                     last_dut_status[dut_name] = dut_status[dut_name]
                     dut_status[dut_name]['status'] = 'idle'
@@ -221,11 +259,27 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
 
                     report_progress('testing', dut_status, sequence=sequence)
 
+            for dut_name, dut_value in duts.items():
+                if not dut_value:
+                    continue
+                if 'failed_step' in failed_steps[dut_name]:
+                    dut_status[dut_name]['test_status'] = 'fail'
+                    dut_status[dut_name]['failed_step'] = failed_steps[dut_name]['failed_step']
+
+                else:
+                    dut_status[dut_name]['test_status'] = 'pass'
+
+                last_dut_status[dut_name] = dut_status[dut_name]
+
             report_progress(
                 'finalize', dut_status, overall_result=overall_result, sequence=sequence
             )
             common_definitions.finalize_test(overall_result, duts, common_definitions.INSTRUMENTS)
 
+            results["end_time"] = datetime.datetime.now()
+
+            results["duration_s"] = (results["end_time"] - results["start_time"]).total_seconds()
+
             results["Overall result"] = overall_result
 
         except exceptions.Error as e:
@@ -239,7 +293,9 @@ def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue):
 
         report_progress("Create test report", sequence=sequence)
         if not test_control['report_off']:
-            common_definitions.create_report(json.dumps(results), duts)
+            common_definitions.create_report(
+                json.dumps(results, indent=4, default=str), results, duts
+            )
 
         if test_control['single_run']:
             test_control['terminate'] = True
