@@ -26,13 +26,13 @@ class TestCase(ABC):
         tmp_result = {}
 
         if error:
-            tmp_result['Success'] = {"limit": None, "measurement": error, "result": False}
+            tmp_result['Error'] = {"limit": None, "measurement": error, "result": False}
         elif limits:
             for result in self.results.items():
                 tmp_result[result[0]] = {
                     "limit": inspect.getsource(limits[result[0]]),
                     "measurement": result[1],
-                    "result": limits[result[0]](result[1]),
+                    "result": limits[result[0]](result[1])
                 }
         else:
             tmp_result['Success'] = {"limit": None, "measurement": None, "result": True}
